#!/usr/bin/env python
"""
A pre-forking SCGI server that uses file descriptor passing to off-load
requests to child worker processes.
"""

# created 2001-11-13 NAS

import sys
import socket
import os
import struct
import select
import signal
import errno
import time
from scgi import passfd

# netstring utility functions
def ns_read_size(input):
    size = ""
    while 1:
        c = input.read(1)
        if c == ':':
            break
        elif not c:
            raise IOError, 'short netstring read'
        size = size + c
    return long(size)

def ns_reads(input):
    size = ns_read_size(input)
    data = ""
    while len(data) < size:
        s = input.read(size)
        if not s:
            raise IOError, 'short netstring read'
        data = data + s
    if input.read(1) != ',':
        raise IOError, 'missing netstring terminator'
    return data


class SCGIHandler:

    # Subclasses should override the handle_connection method.

    def __init__(self, parent_fd, ready_pipe):
        self.parent_fd = parent_fd
        self.ready_pipe = ready_pipe

    def serve(self):
        while 1:
            try:
                fd = passfd.recvfd(self.parent_fd)
            except IOError:
                # parent probably exited
                raise SystemExit
            conn = socket.fromfd(fd, socket.AF_INET, socket.SOCK_STREAM)
            os.close(fd)
            self.handle_connection(conn)
            # tell the parent that we are ready for more work
            os.write(self.ready_pipe, struct.pack("l", os.getpid()))
        

    def read_env(self, input):
        headers = ns_reads(input)
        items = headers.split("\0")
        items = items[:-1]
        assert len(items) % 2 == 0, "malformed headers"
        env = {}
        for i in range(0, len(items), 2):
            env[items[i]] = items[i+1]
        return env

    def handle_connection(self, conn):
        input = conn.makefile("r")
        output = conn.makefile("w")

        env = self.read_env(input)
        output.write("Content-Type: text/plain\r\n")
        output.write("\r\n")
        for k, v in env.items():
            output.write("%s: %r\n" % (k, v))

        output.close()
        input.close()
        conn.close()
        

class SCGIServer:

    DEFAULT_PORT = 4000

    def __init__(self, handler_class=SCGIHandler, host="", port=DEFAULT_PORT,
                 max_children=5):
        self.handler_class = handler_class
        self.host = host
        self.port = port
        self.max_children = max_children
        self.children = {} # { pid : fd }
        self.busy_children = {} # { pid : 1 }
        self.kill_children = {} # { pid : 1 }
        (self.ready_pipe_read, self.ready_pipe_write) = os.pipe()
        self.spawn_child()


    def spawn_child(self, conn=None):
        parent_fd, child_fd = passfd.socketpair(socket.AF_UNIX,
                                                socket.SOCK_STREAM)
        pid = os.fork()
        if pid == 0:
            if conn:
                conn.close() # in the midst of handling a request, close
                             # the connection in the child 
            os.close(child_fd)
            self.handler_class(parent_fd, self.ready_pipe_write).serve()
            sys.exit(0)
        else:
            os.close(parent_fd)
            self.children[pid] = child_fd

    def find_ready_child(self, conn=None):
        """Find an available child process to handle a request.  Return the fd
        connected child (used for passing the request fd).  This method blocks
        if all the children are busy and we have reached the max_children
        limit."""

        while self.busy_children:
            # find which children are finished processing a request
            if len(self.children) == self.max_children:
                # block
                r, w, e = select.select([self.ready_pipe_read], [], [])
            else:
                r, w, e = select.select([self.ready_pipe_read], [], [], 0)
            if not r:
                break
            (pid, ) = struct.unpack("l", os.read(self.ready_pipe_read, 4))
            del self.busy_children[pid]

        # kill off children (scheduled after getting a HUP signal)
        if self.kill_children:
            for pid in self.kill_children.keys():
                os.kill(pid, signal.SIGTERM)
                self.busy_children[pid] = 1
            self.kill_children.clear()
            
        # reap any dead children
        while self.children:
            (pid, status) = os.waitpid(-1, os.WNOHANG)
            if pid <= 0:
                break
            os.close(self.children[pid])
            del self.children[pid]
            try:
                del self.busy_children[pid]
            except KeyError:
                pass

        if len(self.busy_children) == len(self.children):
            # if we get here and all children are busy then we haven't
            # reached max_children yet
            assert len(self.children) < self.max_children
            self.spawn_child(conn)

        # find the first non-busy child
        for pid, fd in self.children.items():
            if not self.busy_children.has_key(pid):
                self.busy_children[pid] = 1
                return fd


    def process_hup(self, signum, frame):
        # schedule termination of children
        for pid in self.children.keys():
            self.kill_children[pid] = 1

    def serve(self):
        signal.signal(signal.SIGHUP, self.process_hup)
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind((self.host, self.port))
        s.listen(10)
        while 1:
            while 1: # restart interrupted system calls
                try:
                    conn, addr = s.accept()
                except socket.error, exc:
                    if exc[0] != errno.EINTR:
                        raise
                else:
                    break
            child_fd = self.find_ready_child(conn)
            passfd.sendfd(child_fd, conn.fileno())
            conn.close()


def main():
    if len(sys.argv) == 2:
        port = int(sys.argv[1])
    else:
        port = SCGIServer.DEFAULT_PORT
    SCGIServer(port=port).serve()

if __name__ == "__main__":
    main()
